--[[
Interface: 1.6.0.0 b9173

Copyright (C) GtX (Andy), 2019

Author: GtX | Andy
Date: 15.12.2019
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Free for use in other mods - no permission needed.
No changes are to be made to this script without permission from GtX | Andy

Frei verwendbar - keine erlaubnis nötig
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


BunkerSiloDisplay = {}

BunkerSiloDisplay.TARGET_MARKER = nil
BunkerSiloDisplay.MAX_CONNECTION_DISTANCE = 15

BunkerSiloDisplay.COLOURS = {
    RED = {1, 0, 0, 1},
    GREEN = {0, 1, 0, 1},
    WHITE = {0.8, 0.8, 0.8, 1},
    YELLOW = {1, 1, 0, 1},
    ORANGE = {0.9, 0.1, 0, 1},
    GREY = {0.3, 0.3, 0.3, 1},
    BLACK = {0.05, 0.05, 0.05, 1},
    LIGHT_RED = {1, 0.5, 0.25, 1},
    LIGHT_GREEN = {0.05, 0.15, 0.05, 1},
    DARK_GREEN = {0, 0.2, 0, 1}
}

local BunkerSiloDisplay_mt = Class(BunkerSiloDisplay, Placeable)
InitObjectClass(BunkerSiloDisplay, "BunkerSiloDisplay")

function BunkerSiloDisplay:new(isServer, isClient, customMt)
    local self = Placeable:new(isServer, isClient, customMt or BunkerSiloDisplay_mt)

    self.bunkerSiloInRange = nil
    self.bunkerSiloPosition = nil
    self.connectionMessage = nil

    self.displays = {}
    self.previewObjectToNodes = {}

    self.currentFillLevel = 0
    self.currentFermentingPercent = 0
    self.currentCompactedPercent = 0

    self.displayConfigurations = {}
    self.numDisplayConfigurations = 0
    self.activeConfigId = 0

    self.objectActivated = false

    self.debugConnection = false
    self.isBunkerSiloDisplay = true

    self.connectionLostText = "Bunker silo not be found!"

    registerObjectClassName(self, "BunkerSiloDisplay")

    return self
end

function BunkerSiloDisplay:load(xmlFilename, x, y, z, rx, ry, rz, initRandom)
    if not BunkerSiloDisplay:superClass().load(self, xmlFilename, x, y, z, rx, ry, rz, initRandom) then
        return false
    end

    local xmlFile = loadXMLFile("TempXML", xmlFilename)

    self:loadDisplay(xmlFile, "fillLevel", 8, "currentFillLevel")
    self:loadDisplay(xmlFile, "fermentingPercent", 3, "currentFermentingPercent")
    self:loadDisplay(xmlFile, "compactedPercent", 3, "currentCompactedPercent")

    local iconNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.bunkerSiloDisplay.icon#node"))
    if iconNode ~= nil then
        self.icon = {
            node = iconNode,
            originalMaterial = getMaterial(iconNode, 0)
        }

        self.icon.materialTypeName = string.upper(Utils.getNoNil(getXMLString(xmlFile, "placeable.bunkerSiloDisplay.icon#materialTypeName"), "ICON"))
        self.icon.parameterName = Utils.getNoNil(getXMLString(xmlFile, "placeable.bunkerSiloDisplay.icon#materialShaderParameterName"), "iconColor")
        self.icon.parameters = Utils.getNoNil(self:getDisplayColour(getXMLString(xmlFile, "placeable.bunkerSiloDisplay.icon#materialShaderParameter")), BunkerSiloDisplay.COLOURS.LIGHT_RED)

        self.icon.originalIsBackup = Utils.getNoNil(getXMLBool(xmlFile, "placeable.bunkerSiloDisplay.icon#originalMaterialIsBackup"), false)

        setVisibility(iconNode, false)
    end

    local i = 0
    while true do
        local languageKey = string.format("placeable.bunkerSiloDisplay.languageNodes.language(%d)", i)
        if not hasXMLProperty(xmlFile, languageKey) then
            break
        end

        local node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, languageKey .. "#node"))
        local short = getXMLString(xmlFile, languageKey .. "#short")

        if node ~= nil and short ~= nil then
            if self.languageNodes == nil then
                self.languageNodes = {}
            end

            if self.languageNodes[short] == nil then
                self.languageNodes[short] = node
            else
                g_logManager:xmlWarning(xmlFilename, "Duplicate language given at '%s'. This should be removed.", languageKey)
            end

            setVisibility(node, false)
        end

        i = i + 1
    end

    local interactionTrigger = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.bunkerSiloDisplay.configurations#interactionTrigger"))

    if interactionTrigger ~= nil then
        local colMask = getCollisionMask(interactionTrigger)

        if bitAND(Player.COLLISIONMASK_TRIGGER, colMask) ~= 0 then
            addTrigger(interactionTrigger, "interactionTriggerCallback", self)

            self.interactionTrigger = interactionTrigger

            self.activateText = g_i18n:getText("shop_configuration")
            self.configurationText = self.activateText .. ":  %s"

            i = 0
            while true do
                local configurationKey = string.format("placeable.bunkerSiloDisplay.configurations.configuration(%d)", i)
                if not hasXMLProperty(xmlFile, configurationKey) then
                    break
                end

                local name = getXMLString(xmlFile, configurationKey .. "#name")
                if name ~= nil then
                    name = g_i18n:convertText(name, self.customEnvironment)

                    local objects = {}
                    ObjectChangeUtil.loadObjectChangeFromXML(xmlFile, configurationKey, objects, self.nodeId, self)

                    table.insert(self.displayConfigurations, {name = name, objects = objects})
                end

                i = i + 1
            end

            self.numDisplayConfigurations = #self.displayConfigurations

            self.defualtConfigId = MathUtil.clamp(Utils.getNoNil(getXMLInt(xmlFile, "placeable.bunkerSiloDisplay.configurations#default"), 1), 1, self.numDisplayConfigurations)
            self:setActiveDisplayConfiguration(self.defualtConfigId, true)
        else
            g_logManager:xmlWarning(xmlFilename, "Invalid collision mask for interaction trigger 'placeable.bunkerSiloDisplay#interactionTrigger'. Bit 20 needs to be set!")
        end
    end

    -- Helps the user get the correct rotation of the display.
    self.placementMarker = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.bunkerSiloDisplay.placementMarker#node"))
    if self.placementMarker ~= nil then
        setVisibility(self.placementMarker, true)
    end

    self.debugConnection = Utils.getNoNil(getXMLBool(xmlFile, "placeable.bunkerSiloDisplay#debugConnection"), self.debugConnection)

    delete(xmlFile)

    return true
end

function BunkerSiloDisplay:loadDisplay(xmlFile, displayType, minElements, valueName)
    local key = "placeable.bunkerSiloDisplay." .. displayType
    local baseNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#baseNode"))

    if baseNode ~= nil then
        local numChildren = getNumOfChildren(baseNode)

        if numChildren >= minElements then
            display = {
                baseNode = baseNode,
                valueName = valueName,
                typeName = displayType,
                precision = getXMLInt(xmlFile, key .. "#precision") or 0,
                showZero = Utils.getNoNil(getXMLBool(xmlFile, key .. "#showZero"), true)
            }

            local colourStr = Utils.getNoNil(self:getDisplayColour(getXMLString(xmlFile, key .. "#color")), BunkerSiloDisplay.COLOURS.LIGHT_RED)

            for element, _ in pairs(I3DUtil.getNodesByShaderParam(baseNode, "numberColor")) do
                setShaderParameter(element, "numberColor", colourStr[1], colourStr[2], colourStr[3], 1, false)
            end

            table.insert(self.displays, display)
        else
            g_logManager:xmlWarning(self.configFileName, "Number of children for node '%s#baseNode' should be %d or greater, only %d children found!", key, minElements, numChildren)
        end
    end
end

function BunkerSiloDisplay:finalizePlacement()
    BunkerSiloDisplay:superClass().finalizePlacement(self)

    if self.placementMarker ~= nil then
        setVisibility(self.placementMarker, false)
        self.placementMarker = nil -- No longer needed!
    end

    if g_currentMission.gameStarted then
        self:attachToBunkerSilo()
    else
        g_currentMission:registerObjectToCallOnMissionStart(self)
    end

    if self.languageNodes ~= nil then
        local languageNode = self.languageNodes["en"]

        if languageNode ~= nil then
            if self.languageNodes[g_languageShort] ~= nil then
                languageNode = self.languageNodes[g_languageShort]
            end

            setVisibility(languageNode, true)
        else
            g_logManager:xmlWarning(self.configFileName, "English (en) language was not found at 'placeable.bunkerSiloDisplay.languageNodes'! This must be set as a backup for other languages not given.")
        end
    end

    if g_i18n:hasText("warning_bunkerSiloConnectionLost", self.customEnvironment) then
        self.connectionLostText = g_i18n:getText("warning_bunkerSiloConnectionLost", self.customEnvironment)
    end
end

function BunkerSiloDisplay:delete()
    self:resetDisplay()

    if self.interactionTrigger ~= nil then
        g_currentMission:removeActivatableObject(self)
        removeTrigger(self.interactionTrigger)

        self.objectActivated = false
        self.interactionTrigger = nil
    end

    if self:hasConnectedBunker() then
        self.bunkerSiloInRange:removeDeleteListener(self, "onBunkerSiloDelete")

        if self.bunkerSiloInRange.bunkerSiloDisplays ~= nil then
            self.bunkerSiloInRange.bunkerSiloDisplays[self] = nil
        end
    end

    unregisterObjectClassName(self)
    BunkerSiloDisplay:superClass().delete(self)
end

function BunkerSiloDisplay:readStream(streamId, connection)
    BunkerSiloDisplay:superClass().readStream(self, streamId, connection)

    if connection:getIsServer() then
        if streamReadBool(streamId) then
            local x = streamReadFloat32(streamId)
            local y = streamReadFloat32(streamId)
            local z = streamReadFloat32(streamId)

            self.bunkerSiloPosition = {
                MathUtil.round(x, 4),
                MathUtil.round(y, 4),
                MathUtil.round(z, 4)
            }
        end

        if self:hasConfigurations() then
            local configId = streamReadUInt8(streamId)
            self:setActiveDisplayConfiguration(configId, true)
        end
    end
end

function BunkerSiloDisplay:writeStream(streamId, connection)
    BunkerSiloDisplay:superClass().writeStream(self, streamId, connection)

    if not connection:getIsServer() then
        if streamWriteBool(streamId, self:hasConnectedBunker() and self.bunkerSiloPosition ~= nil) then
            streamWriteFloat32(streamId, self.bunkerSiloPosition[1])
            streamWriteFloat32(streamId, self.bunkerSiloPosition[2])
            streamWriteFloat32(streamId, self.bunkerSiloPosition[3])
        end

        if self:hasConfigurations() then
            streamWriteUInt8(streamId, self.activeConfigId)
        end
    end
end

function BunkerSiloDisplay:saveToXMLFile(xmlFile, key, usedModNames)
    BunkerSiloDisplay:superClass().saveToXMLFile(self, xmlFile, key, usedModNames)

    if self:hasConnectedBunker() and self.bunkerSiloPosition ~= nil then
        local x, y, z = unpack(self.bunkerSiloPosition)
        setXMLString(xmlFile, key .. ".bunkerSiloDisplay#connection", string.format("%.4f %.4f %.4f", x, y, z))
    end

    if self:hasConfigurations() then
        setXMLInt(xmlFile, key .. ".bunkerSiloDisplay#activeConfigId", self.activeConfigId)
    end
end

function BunkerSiloDisplay:loadFromXMLFile(xmlFile, key, resetVehicles)
    if not BunkerSiloDisplay:superClass().loadFromXMLFile(self, xmlFile, key, resetVehicles) then
        return false
    end

    local x, y, z = StringUtil.getVectorFromString(getXMLString(xmlFile, key .. ".bunkerSiloDisplay#connection"))

    if x ~= nil and y ~= nil and z ~= nil then
        self.bunkerSiloPosition = {x, y, z}
    end

    self:setActiveDisplayConfiguration(getXMLInt(xmlFile, key .. ".bunkerSiloDisplay#activeConfigId"), true)

    return true
end

function BunkerSiloDisplay:update(dt)
    if self.connectionMessage ~= nil and self.objectActivated and g_currentMission.controlPlayer then
        g_currentMission:addExtraPrintText(self.connectionMessage)
        self:raiseActive()
    end
end

function BunkerSiloDisplay:attachToBunkerSilo()
    if not self:hasConnectedBunker() then
        self.bunkerSiloInRange, _ = self:getBunkerSiloInRange(self:getOwnerFarmId(), false)
    end

    if self:hasConnectedBunker() then
        if self.bunkerSiloPosition == nil then
            self.bunkerSiloPosition = {self:getBunkerSiloPosition(self.bunkerSiloInRange)}
        end

        self.bunkerSiloInRange:addDeleteListener(self, "onBunkerSiloDelete")

        if self.bunkerSiloInRange.bunkerSiloDisplays == nil then
            self.bunkerSiloInRange.bunkerSiloDisplays = {}

            if self.bunkerSiloInRange.bunkerSiloDisplayInjected == nil then
                self.bunkerSiloInRange.bunkerSiloDisplayInjected = true

                if self.isServer then
                    if g_dedicatedServerInfo == nil then
                        self.bunkerSiloInRange.raiseDirtyFlags = Utils.appendedFunction(self.bunkerSiloInRange.raiseDirtyFlags, function(bunkerSilo, flag)
                            if bunkerSilo.bunkerSiloDisplays ~= nil then
                                for _, bunkerSiloDisplay in pairs (bunkerSilo.bunkerSiloDisplays) do
                                    bunkerSiloDisplay:updateDisplays(bunkerSilo)
                                end
                            end
                        end)
                    end
                else
                    self.bunkerSiloInRange.readUpdateStream = Utils.appendedFunction(self.bunkerSiloInRange.readUpdateStream, function(bunkerSilo, streamId, timestamp, connection)
                        if connection:getIsServer() and bunkerSilo.bunkerSiloDisplays ~= nil then
                            for _, bunkerSiloDisplay in pairs (bunkerSilo.bunkerSiloDisplays) do
                                bunkerSiloDisplay:updateDisplays(bunkerSilo)
                            end
                        end
                    end)
                end
            end
        end

        self.bunkerSiloInRange.bunkerSiloDisplays[self] = self
        self:updateDisplays(self.bunkerSiloInRange)

        self:setConnectionMessage(nil)
    else
        if self.connectionMessage == nil then
            self:setConnectionMessage(self.connectionLostText)
        end
    end

    self:resetDisplay()
end

function BunkerSiloDisplay:resetDisplay()
    self.bunkerSiloBuffer = nil
    self.connectionPointsBuffer = nil

    self:setTargetMarker(nil, true)
end

function BunkerSiloDisplay:setTargetMarker(areaData, isReset)
    local marker = BunkerSiloDisplay.TARGET_MARKER

    if marker ~= nil then
        if areaData ~= nil then
            local xz = math.min(areaData.width * 0.09, 1)

            setScale(marker, xz, 1, xz)
            setTranslation(marker, areaData.centreX, areaData.centreY, areaData.centreZ)

            setVisibility(marker, true)
        else
            setVisibility(marker, false)

            if isReset then
                setScale(marker, 1, 1, 1)
                setTranslation(marker, 0, 0, 0)
            end
        end
    end
end

function BunkerSiloDisplay:updateDisplays(object)
    if object == nil or object.bunkerSiloDirtyFlag == nil or object ~= self.bunkerSiloInRange then
        return
    end

    self.currentFillLevel = math.max(object.fillLevel, 0)

    if object.state == BunkerSilo.STATE_FILL then
        self.currentFermentingPercent = 0
        self.currentCompactedPercent = object.compactedPercent
    else
        self.currentFermentingPercent = object.fermentingPercent
        self.currentCompactedPercent = 100
    end

    for _, display in pairs (self.displays) do
        I3DUtil.setNumberShaderByValue(display.baseNode, self[display.valueName], display.precision, display.showZero)
    end

    if self.icon ~= nil then
        local material
        local fillTypeIndex = object.inputFillType

        if object.state == BunkerSilo.STATE_FERMENTED or object.state == BunkerSilo.STATE_DRAIN then
            fillTypeIndex = object.outputFillType
        end

        if fillTypeIndex ~= nil and fillTypeIndex ~= self.icon.fillTypeIndex then
            self.icon.fillTypeIndex = fillTypeIndex

            material = g_materialManager:getMaterial(fillTypeIndex, self.icon.materialTypeName, 1)

            if material == nil and self.icon.originalIsBackup then
                material = self.icon.originalMaterial
            end

            if material ~= nil then
                local x, y, z, w = unpack(self.icon.parameters)

                setMaterial(self.icon.node, material, 0)
                setShaderParameter(self.icon.node, self.icon.parameterName, x, y, z, w, false)

                setVisibility(self.icon.node, true)
            else
                setVisibility(self.icon.node, false)
            end
        end
    end
end

function BunkerSiloDisplay:setActiveDisplayConfiguration(configId, noEventSend)
    if self:hasConfigurations() then
        if configId ~= nil and self.displayConfigurations[configId] ~= nil then
            BunkerSiloDisplayConfigEvent.sendEvent(self, configId, noEventSend)

            self.activateText = string.format(self.configurationText, self.displayConfigurations[configId].name)

            if self.activeConfigId ~= configId then
                self.activeConfigId = configId

                for i = 1, self.numDisplayConfigurations do
                    ObjectChangeUtil.setObjectChanges(self.displayConfigurations[i].objects, self.activeConfigId == i)
                end
            end
        end
    end
end

function BunkerSiloDisplay:getCanBePlacedAt(x, y, z, distance, farmId)
    local canBePlaced = BunkerSiloDisplay:superClass().getCanBePlacedAt(self, x, y, z, distance, farmId)
    local bunkerSiloInRange

    if canBePlaced then
        local ox, oy, oz = getTranslation(self.nodeId)
        setTranslation(self.nodeId, x, y, z)

        bunkerSiloInRange, canBePlaced = self:getBunkerSiloInRange(farmId, true)
        setTranslation(self.nodeId, ox, oy, oz)
    end

    self.bunkerSiloInRange = bunkerSiloInRange

    return canBePlaced
end

function BunkerSiloDisplay:getBunkerSiloInRange(farmId, isPlacement)
    local bunkerSiloInRange, hasConnection = nil, false

    if self.nodeId ~= 0 then
        local bufferIndex, lastDistance, connectionPointId = nil, math.huge, 1
        local x, y, z = getWorldTranslation(self.nodeId)

        if self.bunkerSiloBuffer == nil then
            self.bunkerSiloBuffer = {}
            self.connectionPointsBuffer = {0, 0, 0, 0}

            for _, placeable in pairs(g_currentMission.placeables) do
                if placeable:getOwnerFarmId() == farmId and placeable.bunkerSilos ~= nil then

                    -- Only connect to Bunker Silos, BGA Bunker Silos or mod script placeables that allow connection using 'allowBunkerSiloDisplays = true'.
                    if placeable:isa(BunkerSiloPlaceable) or placeable:isa(BgaPlaceable) or placeable.allowBunkerSiloDisplays then
                        for _, bunkerSilo in ipairs(placeable.bunkerSilos) do
                            if self.bunkerSiloPosition == nil then
                                local area = bunkerSilo.bunkerSiloArea

                                local widthX, _, widthZ = localToLocal(area.start, area.width, 0, 0, 0)
                                local heightX, _, heightZ = localToLocal(area.start, area.height, 0, 0, 0)

                                local _, ry, _ = getWorldRotation(area.start)

                                local data = {
                                    yRot = 0,
                                    bunkerSilo = bunkerSilo,
                                    width = MathUtil.vector2Length(widthX, widthZ),
                                    height = MathUtil.vector2Length(heightX, heightZ)
                                }

                                -- The 'area' on some mods is set up incorrectly! Try and correct for placement rendering here only.

                                local offsetX1 = -data.width
                                local offsetZ1 = 0

                                local offsetX2 = -data.width
                                local offsetZ2 = -data.height

                                if heightX == 0 then
                                    data.yRot = ry

                                    if widthX < 0 then
                                        offsetX1 = data.width
                                        offsetX2 = offsetX1
                                    end

                                    if heightZ < 0 then
                                        offsetZ2 = data.height
                                    end
                                else
                                    offsetX1 = 0
                                    data.yRot = ry - (math.pi * 0.5)

                                    if heightX > 0 then
                                        offsetZ1 = -data.width
                                        offsetX2 = -data.height
                                    else
                                        offsetZ1 = data.width
                                        offsetX2 = data.height
                                    end

                                    if widthZ > 0 then
                                        offsetZ2 = -data.width
                                        offsetZ1 = offsetZ2
                                    else
                                        offsetZ2 = data.width
                                        offsetZ1 = offsetZ2
                                    end
                                end

                                data.startX, data.startY, data.startZ = localToWorld(area.start, 0, 0, 0)
                                data.widthSX, data.widthSY, data.widthSZ = localToWorld(area.width, 0, 0, 0)
                                data.widthHX, data.widthHY, data.widthHZ = localToWorld(area.height, offsetX1, 0, offsetZ1)
                                data.heightX, data.heightY, data.heightZ = localToWorld(area.height, 0, 0, 0)

                                data.centreX, data.centreY, data.centreZ = localToWorld(area.start, offsetX2 * 0.5, 0, offsetZ2 * 0.5)

                                table.insert(self.bunkerSiloBuffer, data)
                            else
                                local posX, posY, posZ = self:getBunkerSiloPosition(bunkerSilo)

                                if posX == self.bunkerSiloPosition[1] and posY == self.bunkerSiloPosition[2] and posZ == self.bunkerSiloPosition[3] then
                                    return bunkerSilo, true
                                end
                            end
                        end
                    end
                end
            end
        end

        for index, bunkerSiloData in ipairs(self.bunkerSiloBuffer) do
            if bunkerSiloData.bunkerSilo ~= nil then
                self.connectionPointsBuffer[1] = MathUtil.vector3Length(x - bunkerSiloData.startX, y - bunkerSiloData.startY, z - bunkerSiloData.startZ)
                self.connectionPointsBuffer[2] = MathUtil.vector3Length(x - bunkerSiloData.widthSX, y - bunkerSiloData.widthSY, z - bunkerSiloData.widthSZ)
                self.connectionPointsBuffer[3] = MathUtil.vector3Length(x - bunkerSiloData.widthHX, y - bunkerSiloData.widthHY, z - bunkerSiloData.widthHZ)
                self.connectionPointsBuffer[4] = MathUtil.vector3Length(x - bunkerSiloData.heightX, y - bunkerSiloData.heightY, z - bunkerSiloData.heightZ)

                for id, distance in ipairs (self.connectionPointsBuffer) do
                    if distance < BunkerSiloDisplay.MAX_CONNECTION_DISTANCE and distance < lastDistance then
                        bunkerSiloInRange = bunkerSiloData.bunkerSilo
                        connectionPointId = id
                        lastDistance = distance
                        hasConnection = true
                        bufferIndex = index
                    end
                end
            end
        end

        if isPlacement then
            if hasConnection then
                if bunkerSiloInRange ~= self.bunkerSiloInRange then
                    self:setTargetMarker(self.bunkerSiloBuffer[bufferIndex], false)
                end
            else
                self:setTargetMarker(nil, false)
            end

            if self.debugConnection then
                if hasConnection then
                    local startNode = bunkerSiloInRange.bunkerSiloArea.start
                    local areaData = self.bunkerSiloBuffer[bufferIndex]

                    if areaData ~= nil and startNode ~= nil then
                        local connectedDisplays = 0

                        local upX, upY, upZ = localDirectionToWorld(startNode, 0, 1, 0)
                        local dirX, dirY, dirZ = localDirectionToWorld(startNode, 0, 0, 1)

                        if bunkerSiloInRange.bunkerSiloDisplays ~= nil then
                            for _ in pairs(bunkerSiloInRange.bunkerSiloDisplays) do
                                connectedDisplays = connectedDisplays + 1
                            end
                        end

                        if connectionPointId == 1 then
                            DebugUtil.drawDebugGizmoAtWorldPos(areaData.startX, areaData.startY, areaData.startZ, dirX, dirY, dirZ, upX, upY, upZ, "Connection Point", false)
                        elseif connectionPointId == 2 then
                            DebugUtil.drawDebugGizmoAtWorldPos(areaData.widthSX, areaData.widthSY, areaData.widthSZ, dirX, dirY, dirZ, upX, upY, upZ, "Connection Point", false)
                        elseif connectionPointId == 3 then
                            DebugUtil.drawDebugGizmoAtWorldPos(areaData.widthHX, areaData.widthHY, areaData.widthHZ, dirX, dirY, dirZ, upX, upY, upZ, "Connection Point", false)
                        else
                            DebugUtil.drawDebugGizmoAtWorldPos(areaData.heightX, areaData.heightY, areaData.heightZ, dirX, dirY, dirZ, upX, upY, upZ, "Connection Point", false)
                        end

                        Utils.renderTextAtWorldPosition(areaData.centreX, areaData.centreY, areaData.centreZ, string.format("Connected Displays:  %d", connectedDisplays), getCorrectTextSize(0.012), 0)
                    end
                end

                setTextColor(1, 1, 0.25, 1)
                setTextAlignment(RenderText.ALIGN_CENTER)

                renderText(0.5, 0.025, getCorrectTextSize(0.02), string.format("Bunker Silo Buffer:  %d", #self.bunkerSiloBuffer))

                setTextColor(1, 1, 1, 1)
                setTextAlignment(RenderText.ALIGN_LEFT)
            end
        end
    end

    return bunkerSiloInRange, hasConnection
end

function BunkerSiloDisplay:getBunkerSiloPosition(bunkerSilo)
    local posX, posY, posZ = getWorldTranslation(bunkerSilo.bunkerSiloArea.start)
    return MathUtil.round(posX, 4), MathUtil.round(posY, 4), MathUtil.round(posZ, 4)
end

function BunkerSiloDisplay:getDisplayColour(colourStr)
    if colourStr == nil then
        return nil
    end

    local colourUpper = colourStr:upper()
    if BunkerSiloDisplay.COLOURS[colourUpper] ~= nil then
        return BunkerSiloDisplay.COLOURS[colourUpper]
    end

    local brandColor = g_brandColorManager:getBrandColorByName(colourStr)
    if brandColor ~= nil then
        return brandColor
    end

    local vector = StringUtil.getVectorNFromString(colourStr, 4)
    if vector ~= nil then
        return vector
    end
end

function BunkerSiloDisplay:setConnectionMessage(connectionMessage)
    self.connectionMessage = connectionMessage
    self:raiseActive()
end

function BunkerSiloDisplay:hasConnectedBunker()
    return self.bunkerSiloInRange ~= nil
end

function BunkerSiloDisplay:getConnectedBunker()
    return self.bunkerSiloInRange
end

function BunkerSiloDisplay:hasConfigurations()
    return self.numDisplayConfigurations > 0
end

function BunkerSiloDisplay:interactionTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
    if (onEnter or onLeave) and g_currentMission.player ~= nil and otherActorId == g_currentMission.player.rootNode then
        if onEnter then
            if not self.objectActivated then
                if self.ownerFarmId == nil or self.ownerFarmId == AccessHandler.EVERYONE or
                    g_currentMission.accessHandler:canFarmAccessOtherId(g_currentMission:getFarmId(), self.ownerFarmId) then

                    g_currentMission:addActivatableObject(self)
                    self.objectActivated = true
                end
            end
        elseif self.objectActivated then
            g_currentMission:removeActivatableObject(self)
            self.objectActivated = false
        end

        if self.connectionMessage ~= nil then
            self:raiseActive()
        end
    end
end

function BunkerSiloDisplay:onMissionStarted()
    self:attachToBunkerSilo()
end

function BunkerSiloDisplay:onBunkerSiloDelete(object)
    if object.bunkerSiloDisplays ~= nil then
        object.bunkerSiloDisplays[self] = nil
    end

    for _, display in pairs (self.displays) do
        I3DUtil.setNumberShaderByValue(display.baseNode, 0, display.precision, false)
    end

    if self.icon ~= nil then
        setVisibility(self.icon.node, false)
        self.icon.fillTypeIndex = nil
    end

    self.bunkerSiloInRange = nil
    -- self.bunkerSiloPosition = nil

    if self.connectionMessage == nil then
        self:setConnectionMessage(self.connectionLostText)
    end
end

function BunkerSiloDisplay:getIsActivatable()
    return g_currentMission.controlPlayer
end

function BunkerSiloDisplay:drawActivate()
end

function BunkerSiloDisplay:shouldRemoveActivatable()
    return false
end

function BunkerSiloDisplay:onActivateObject()
    local configId = self.activeConfigId + 1

    if configId > self.numDisplayConfigurations then
        configId = 1
    end

    self:setActiveDisplayConfiguration(configId)
end
